#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This module contains scatter plot routines for Aerocom data.
"""
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from pyaerocom import const, logger, change_verbosity

# text positions for the annotations
xypos=[]
xypos.append((.01, 0.95))
xypos.append((0.01, 0.90))
xypos.append((0.3, 0.90))
xypos.append((0.01, 0.86))
xypos.append((0.3, 0.86))
xypos.append((0.01, 0.82))
xypos.append((0.3, 0.82))
xypos.append((0.01, 0.78))
xypos.append((0.3, 0.78))
xypos.append((0.8, 0.1))
xypos.append((0.8, 0.06))


def plotscatter(model_name, model_data=None, obs_data=None, opts=None, 
                verbose=True):
    """Method to plot scatterplots

    Todo
    ----

    Complete docstring, review code

    """
    if verbose:
        change_verbosity(new_level='debug')
        
    plt_name = 'SCATTERLOG'
    var_to_run = opts['VariablesToRun'][0]
    
    # global settings (including plot settings) for variable
    VARS = const.VARS[var_to_run]
    
    obs_network_name = opts['ObsNetworkName'][0]
    obs_data_as_series = obs_data.to_timeseries(start_date=opts['StartDate'], 
                                                end_date=opts['EndDate'], 
                                                freq='D')
    obs_lats = [obs_data_as_series[i]['latitude'] for i in range(len(obs_data_as_series))]
    obs_lons = [obs_data_as_series[i]['longitude'] for i in range(len(obs_data_as_series))]
    obs_names = [obs_data_as_series[i]['station_name'] for i in range(len(obs_data_as_series))]
    # model_station_data = model_data.interpolate([("latitude", obs_lats), ("longitude", obs_lons)])
    # times_as_dt64 = pa.helpers.cftime_to_datetime64(model_station_data.time)
    # model_data_as_series = pa.helpers.to_time_series_griesie(model_station_data.grid.data, obs_lats, obs_lons,
    #                                                          times_as_dt64, var_name = [var_to_run])

    model_data_as_series = model_data.to_time_series([("latitude", obs_lats), 
                                                      ("longitude", obs_lons)])

    df_time = pd.DataFrame()
    df_points = pd.DataFrame()
    station_no = 0
    for i in range(len(obs_data_as_series)):
        _len = len(obs_data_as_series[i][var_to_run])
        # print('{} length: {}'.format(obs_names[i],_len))
        if _len > 0:
            _nansum = np.nansum(obs_data_as_series[i][var_to_run])
            # _isnan = np.isnan(_nansum)
            # print('{} nansum: {:.3f}'.format(obs_names[i],np.nansum(obs_data_as_series[i][var_to_run])))
            # print('{} isnan: {}'.format(obs_names[i],_isnan))
            if _nansum > np.float_(0.):
                station_no += 1
                # print('{} station_no: {}'.format(obs_names[i],station_no))
            else:
                print('{} removed due to NaNs only'.format(obs_names[i]))
        else:
            continue
        # put obs and model in DataFrame to make them use the same time index
        df_time_temp = pd.DataFrame(obs_data_as_series[i][var_to_run],
                                    columns=[obs_network_name])
        df_points = df_points.append(df_time_temp)
        # df_time_temp[model_name] = model_data_as_series[i][var_to_run]*1.E3
        df_time_temp[model_name] = (model_data_as_series[i][var_to_run] *
                                    VARS['scat_scale_factor'])
        # df_time has now all time steps where either one of the obs or model data have data
        #
        # df_points = df_points.append(pd.DataFrame(np.float_(df_time_temp.values), columns=df_time_temp.columns))
        df_time = df_time.append(pd.DataFrame(df_time_temp, columns=df_time_temp.columns))

    # remove all indices where either one of the data pairs is NaN
    # mainly done to get the number of days right.
    # df_time.corr() gets it right without
    df_time = df_time.dropna(axis=0, how='any')
    df_points = df_points.dropna()
    print('# of measurements: {}'.format(len(df_points)))

    filter_name = 'WORLD-wMOUNTAINS'
    filter_name = 'WORLD'
    time_step_name = 'mALLYEARdaily'
    # OD550_AER_an2008_YEARLY_WORLD_SCATTERLOG_AeronetSunV3Lev2.0.daily.ps.png
    # if df_time[model_name].index[0].year != df_time[model_name].index[-1].year:
    years_covered = df_time[model_name].index[:].year.unique().sort_values()
    if len(years_covered) > 1:
        figname = '{}_{}_an{}-{}_{}_{}_{}_{}.png'.format(model_name,
                                                         var_to_run,years_covered[0],
                                                      years_covered[-1],
                                                      time_step_name, filter_name, plt_name,
                                                      obs_network_name)
        plotname = "{}-{} {}".format(years_covered[0], years_covered[-1], 'daily')
    else:
        figname = '{}_{}_an{}_{}_{}_{}_{}.png'.format(model_name,
                                                      var_to_run,years_covered[0],
                                                      time_step_name, filter_name, plt_name,
                                                      obs_network_name)
        plotname = "{} {}".format(years_covered[0], 'daily')

    logger.info(figname)
    
    mean = df_time.mean()
    correlation_coeff = df_time.corr()
    # IDL: rms=sqrt(total((f_YData-f_Xdata)^2)/n_elements(f_YData))
    #sum = df_time.sum()
    # nmb=total(f_YData-f_Xdata)/total(f_Xdata)*100.
    # c=n_elements(f_YData)
    # f_temp=(f_YData-f_Xdata)/(f_YData+f_Xdata)
    # mnmb=2./c*total(f_temp)*100.
    # fge=2./c*total(abs(f_temp))*100.
    # f_YDatabc=f_YData*(total(f_Xdata,/nan)/total(f_YData,/nan)) ; bias corrected model data
    # rmsbc=sqrt(total((f_YDatabc-f_Xdata)^2)/n_elements(f_YDatabc))
    difference = df_time[model_name] - df_time[obs_network_name]
    num_points = len(df_time)
    rms = np.sqrt(np.nansum(np.power(difference.values, 2)) / num_points)
    nmb = np.sum(difference)/np.sum(df_time[obs_network_name])*100.
    tmp = (df_time[model_name] - df_time[obs_network_name])/(df_time[model_name] + df_time[obs_network_name])
    mnmb = 2./num_points * np.sum(tmp) * 100.
    fge = 2./np.sum(np.abs(tmp)) * 100.

    df_time.plot.scatter(obs_network_name, model_name,
                         loglog=VARS['scat_loglog'],
                         marker='+',
                         color='black')
    # plot the 1 by 1 line
    plt.plot(VARS['scat_xlim'], 
             VARS['scat_ylim'], '-', color='grey')
    plt.axes().set_aspect('equal')

    plt.xlim(VARS['scat_xlim'])
    plt.ylim(VARS['scat_ylim'])
    xypos_index = 0
    var_str = var_to_run + VARS.unit_str
    plt.axes().annotate("{} #: {} # st: {}".format(var_str, 
                        len(df_time), station_no),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=14, color='red')
    xypos_index += 1
    plt.axes().annotate('Obs: {:.3f}'.format(mean[obs_network_name]),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    xypos_index += 1
    plt.axes().annotate('Mod: {:.3f}'.format(mean[model_name]),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    xypos_index += 1
    plt.axes().annotate('NMB: {:.1f}%'.format(nmb),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    xypos_index += 1
    plt.axes().annotate('MNMB: {:.1f}%'.format(mnmb),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    xypos_index += 1
    plt.axes().annotate('R: {:.3f}'.format(correlation_coeff.values[0, 1]),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    xypos_index += 1
    plt.axes().annotate('RMS: {:.3f}'.format(rms),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    xypos_index += 1
    plt.axes().annotate('FGE: {:.3f}'.format(fge),
                        xy=xypos[xypos_index], xycoords='axes fraction', 
                        fontsize=10, color='red')
    # right lower part
    plt.axes().annotate('{}'.format(plotname),
                        xy=xypos[-2], xycoords='axes fraction', ha='center', 
                        fontsize=10, color='black')
    plt.axes().annotate('{}'.format(filter_name),
                        xy=xypos[-1], xycoords='axes fraction', ha='center', 
                        fontsize=10, color='black')

    plt.savefig(figname, dpi=300)
    plt.close()
    




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Module containing helper functions related to iris I/O methods. These contain
reading of Cubes, and some methods to perform quality checks of the data, e.g.

1. checking and correction of time definition
2. number and length of dimension coordinates must match data array
3. Longitude definition from -180 to 180 (corrected if defined on 0 -> 360 intervall)

"""
import os
from datetime import datetime
from numpy import datetime64, asarray, arange
import cf_units

import iris
from iris.experimental.equalise_cubes import equalise_attributes

from pyaerocom import const, logger
from pyaerocom.exceptions import (NetcdfError, VariableDefinitionError,
                                  FileConventionError, 
                                  UnresolvableTimeDefinitionError)

from pyaerocom.helpers import cftime_to_datetime64
from pyaerocom.time_config import TS_TYPE_TO_PANDAS_FREQ
from pyaerocom.io.helpers import add_file_to_log
from pyaerocom.io.fileconventions import FileConventionRead

TSTR_TO_NP_DT = {"hourly"  :  "datetime64[h]",
                 "3hourly" :  "datetime64[3h]",
                 "daily"   :  "datetime64[D]",
                 "monthly" :  "datetime64[M]"}

TSTR_TO_NP_TD = {"hourly"  :  "timedelta64[h]",
                 "3hourly" :  "timedelta64[3h]",
                 "daily"   :  "timedelta64[D]",
                 "monthly" :  "timedelta64[M]"}


TSTR_TO_CF = {"hourly"  :  "hours",
              "3hourly" :  "hours",
              "daily"   :  "days",
              "monthly" :  "days"}

def _load_cubes_custom_multiproc(files, var_name=None, file_convention=None, 
                                 perform_fmt_checks=True, num_proc=None):
    """Like :func:`load_cubes_custom` but faster
    
    Uses multiprocessing module to distribute loading of multiple NetCDF files
    into iris cube.
    
    Parameters
    ----------
    file : list
        netcdf file
    var_name : str
        name of variable to read
    quality_check : bool
        if True, then a quality check of data is performed against the
        information provided in the filename
    file_convention : :obj:`FileConventionRead`, optional
        Aerocom file convention. If provided, then the data content (e.g. 
        dimension definitions) is tested against definition in file name
    perform_fmt_checks : bool
        if True, additional quality checks (and corrections) are (attempted to
        be) performed.
    num_proc : int
        number of jobs 
    
    Returns
    -------
    iris.cube.CubeList
        loaded data as Cube
    """    
    import multiprocessing
    from functools import partial
    if num_proc is None:
        num_proc = multiprocessing.cpu_count() * 2
    func = partial(load_cube_custom, 
                   var_name=var_name, file_convention=file_convention, 
                   perform_fmt_checks=perform_fmt_checks)
    p = multiprocessing.Pool(processes=num_proc)
    return p.map(func, files)

def load_cubes_custom(files, var_name=None, file_convention=None, 
                      perform_fmt_checks=True, **kwargs):
    """Load multiple NetCDF files into CubeList
    
    Parameters
    ----------
    files : list
        netcdf file
    var_name : str
        name of variable to read
    quality_check : bool
        if True, then a quality check of data is performed against the
        information provided in the filename
    file_convention : :obj:`FileConventionRead`, optional
        Aerocom file convention. If provided, then the data content (e.g. 
        dimension definitions) is tested against definition in file name
    perform_fmt_checks : bool
        if True, additional quality checks (and corrections) are (attempted to
        be) performed.
    **kwargs 
        additional keyword args that are parsed to 
        :func:`_load_cubes_custom_multiproc` in case number of input files 
        is larger than 4.
        
    Returns
    -------
    tuple
        2-element tuple containing:
            
            - CubeList, containing loaded cubes
            - list, list of filenames that were successfully loaded 
    """
    from pyaerocom import const
# =============================================================================
#     if len(files) >= 1000:
#         try:
#             cubes =  _load_cubes_custom_multiproc(files, var_name, 
#                                                   file_convention, 
#                                                   perform_fmt_checks,
#                                                   **kwargs)
#             # if this worked, all input files were successfully loaded and the 
#             # function terminates, else, the retrieval is done file by file.
#             # ... Maybe!
#             return (cubes, files)
#         except:
#             pass
# =============================================================================
    cubes = []
    loaded_files = []
    print_where = False
    if len(files) > 10:
        mod = len(files) / 10
        print_where = True
    for i, _file in enumerate(files):
        if print_where and i%mod==0:
            const.print_log.info(os.path.basename(_file))
        try:
            cube = load_cube_custom(_file, var_name,
                                    file_convention=file_convention)
            cubes.append(cube)
            loaded_files.append(_file)
        except Exception as e:
            msg = ("Failed to load {} as Iris cube. Error: {}"
                   .format(_file, repr(e)))
            const.logger.warning(msg)
            
            if const.WRITE_FILEIO_ERR_LOG:
                add_file_to_log(_file, msg)
    return (cubes, loaded_files)

def load_cube_custom(file, var_name=None, file_convention=None, 
                     perform_fmt_checks=None):
    """Load netcdf file as iris.Cube
    
    Parameters
    ----------
    file : str
        netcdf file
    var_name : str
        name of variable to read
    quality_check : bool
        if True, then a quality check of data is performed against the
        information provided in the filename
    file_convention : :obj:`FileConventionRead`, optional
        Aerocom file convention. If provided, then the data content (e.g. 
        dimension definitions) is tested against definition in file name
    perform_fmt_checks : bool
        if True, additional quality checks (and corrections) are (attempted to
        be) performed.
    
    Returns
    -------
    iris.cube.Cube
        loaded data as Cube
    """
    if perform_fmt_checks is None:
        perform_fmt_checks = const.GRID_IO.PERFORM_FMT_CHECKS
    cube_list = iris.load(file)
    
    _num = len(cube_list)
    if _num != 1:
        if _num == 0:
            raise NetcdfError('Data from file {} could not be loaded using iris'
                              .format(file))
        else:
            logger.warning('File {} contains more than one variable'
                           .format(file))
    cube = None
    if var_name is None:
        if not len(cube_list) == 1:
            vars_avail = [c.var_name for c in cube_list]
            raise NetcdfError('Could not load single cube from {}. Please '
                              'specify var_name. Input file contains the '
                              'following variables: {}'.format(file, 
                                                               vars_avail))
        cube = cube_list[0]
        var_name = cube.var_name
    else:
        for c in cube_list:
            if c.var_name == var_name:
                cube = c
                break
    if cube is None:
        raise NetcdfError('Variable {} not available in file {}'.format(var_name, 
                                                                        file))
    if perform_fmt_checks:
        try:
            cube = _check_var_unit_cube(cube)
        except VariableDefinitionError:
            pass
        
        grid_io = const.GRID_IO
        if grid_io.CHECK_TIME_FILENAME:
            try:
                cube = _check_correct_time_dim(cube, file,  file_convention)
            except FileConventionError:
                const.print_log.warning('WARNING: failed to check / validate '
                                        'time dim. using information in '
                                        'filename. Reason: invalid file name '
                                        'convention')
        else:
            logger.warning("WARNING: Automatic check of time "
                           "array in netCDF files is deactivated. "
                           "This may cause problems in case "
                           "the time dimension is not CF conform.") 
        if grid_io.CHECK_DIM_COORDS:
            cube = check_dim_coords_cube(cube)
        
        try:
            if grid_io.DEL_TIME_BOUNDS:
                cube.coord("time").bounds = None
        except:
            logger.warning("Failed to access time coordinate in GriddedData")
            
        if grid_io.SHIFT_LONS:
            cube = check_and_regrid_lons_cube(cube)
    return cube

def check_and_regrid_lons_cube(cube):
    """Checks and corrects for if longitudes of :attr:`grid` are 0 -> 360
    
    Note
    ----
    This method checks if the maximum of the current longitudes array
    exceeds 180. Thus, it is not recommended to use this function after
    subsetting a cube, rather, it should be checked directly when the 
    file is loaded (cf. :func:`load_input`)
    
    Parameters
    ----------
    cube : iris.cube.Cube
        gridded data loaded as iris.Cube
        
    Returns
    -------
    bool
        True, if longitudes were on 0 -> 360 and have been rolled, else
        False
    """
    from pyaerocom import print_log
    try:
        if cube.coord("longitude").points.max() > 180:
            logger.info("Rolling longitudes to -180 -> 180 definition")
            cube = cube.intersection(longitude=(-180, 180))
    except Exception as e:
        print_log.warning('Failed to roll longitudes: {}'.format(repr(e)))
    return cube

def check_dim_coord_names_cube(cube):
    
    from pyaerocom import const
    coords = dict(lon = const.COORDINFO.lon,
                  lat = const.COORDINFO.lat,
                  time = const.COORDINFO.time)

    for coord in cube.dim_coords:
        cv, cs, cn = coord.var_name, coord.standard_name, coord.long_name
        c=None
        if cv in coords:
            c = coords[cv]
        elif cs in coords:
            c = coords[cs]
        elif cn in coords:
            c = coords[cn]
        if c is not None:
            var_name = c.var_name
            std_name = c.standard_name
            lng_name = c.long_name
            if not coord.var_name == var_name:
                const.logger.warning('Invalid var_name {} for coord {} '
                                        'in cube. Overwriting with {}'
                                        .format(coord.standard_name, 
                                                coord.var_name, 
                                                std_name))
                coord.var_name = var_name
            if not coord.standard_name == std_name:
                const.logger.warning('Invalid standard_name {} for coord {} '
                                        'in cube. Overwriting with {}'
                                        .format(coord.standard_name, 
                                                coord.var_name, 
                                                std_name))
                coord.standard_name = std_name
            if not coord.long_name == lng_name:
                const.logger.warning('Invalid long_name {} for coord {} in '
                                        'cube. Overwriting with {}'
                                        .format(coord.long_name, 
                                                coord.var_name, 
                                                lng_name))
                coord.long_name = lng_name
    return cube

def check_dim_coords_cube(cube):
    """Checks, and if necessary and applicable, updates coords names in Cube
    
    Parameters
    ----------
    cube : iris.cube.Cube
        input cube
    
    Returns
    -------
    iris.cube.Cube
        updated or unchanged cube
    """
    cube = check_dim_coord_names_cube(cube)
    return cube


def _check_var_unit_cube(cube):
    var = cube.var_name
    if not var in const.VARS:
        raise VariableDefinitionError('No such pyaerocom default variable: {}'
                                      .format(cube.var_name))
    
    u = cube.units
    if isinstance(u, str):
        u = cf_units.Unit(u)
    if str(const.VARS[var].units) == '1' and u.is_unknown():
        const.print_log.info('Overwriting unit {} in cube {} with value "1"'
                             .format(str(u), var))
        cube.units = cf_units.Unit('1')
    return cube

def check_time_coordOLD(cube, ts_type, year):
    """Method that checks the time coordinate of an iris Cube
    
    This method checks if the time dimension of a cube is accessible and 
    according to the standard (i.e. fully usable). It only checks, and does not
    correct. For the latter, please see :func:`correct_time_coord`.
    
    Parameters
    ----------
    cube : Cube
        cube containing data
    ts_type : str
        temporal resolution of data (e.g. "hourly", "daily"). This information
        is e.g. encrypted in the filename of a NetCDF file and may be 
        accessed using :class:`pyaerocom.io.FileConventionRead`
    year : int
        interger specifying year of observation, e.g. 2017
    
    Returns
    -------
    bool
        True, if time dimension is ok, False if not
    """
    
    ok = True
    test_idx = [0,1,2,7] #7, since last accessible index in a 3hourly dataset of one day is 7
    try:
        try:
            t = cube.coord("time")
        except:
            raise AttributeError("Cube does not contain time dimension")
        if not isinstance(t, iris.coords.DimCoord):
            raise AttributeError("Time is not a DimCoord instance")
        try:
            cftime_to_datetime64(0, cfunit=t.units)
        except:
            raise ValueError("Could not convert time unit string")
        tres_np = TSTR_TO_NP_TD[ts_type]
        conv = TSTR_TO_NP_DT[ts_type]
        base = datetime64("%s-01-01 00:00:00" %year).astype(conv)
        test_datenums = asarray(test_idx)
        ts_nominal = base + test_datenums.astype(tres_np)
        dts_nominal = ts_nominal[1:] - ts_nominal[:-1]
        ts_values = cftime_to_datetime64(t[test_idx].points, cfunit=t.units).astype(conv)
        dts_values = ts_values[1:] - ts_values[:-1]
        if not all(ts_values == ts_nominal):
            raise ValueError("Time match error, nominal dates for test array"
                             "%s (unit=%s): %s\nReceived values after "
                             "conversion: %s"
                             %(test_datenums, t.units.origin,
                               ts_nominal, ts_values))
        elif not all(dts_values == dts_nominal):
            raise ValueError("Time match error, time steps for test array"
                             "%s (unit=%s): %s\nReceived values after "
                             "conversion: %s"
                             %(test_datenums, t.units.origin,
                               dts_nominal, dts_values))
    except Exception as e:
        logger.warning("Invalid time dimension.\n"
                       "Error message: {}".format(repr(e)))
        ok = False
    return ok

def make_datetimeindex(ts_type, year):
    """Create pandas datetime index 
    
    Parameters
    ----------
    ts_type : str
        pyaerocom ts_type
    year : int
        year
        
    Returns
    -------
    pandas.DatetimeIndex
        index object
    """
    import pandas as pd
    start = datetime64("{}-01-01 00:00:00".format(year))
    stop = datetime64("{}-12-31 23:59:59".format(year))
    
    idx = pd.DatetimeIndex(start=start, end=stop, 
                           freq=TS_TYPE_TO_PANDAS_FREQ[ts_type])
    return idx



def _check_correct_time_dim(cube, file, file_convention=None):
    if file_convention is None:
        try:
            file_convention = FileConventionRead(from_file=file)
        except:
            pass
    
    if not isinstance(file_convention, FileConventionRead):
        
        raise FileConventionError('Unknown file convention: {}'
                                  .format(file_convention))
        
    finfo = file_convention.get_info_from_file(file)
    ts_type = finfo['ts_type']
    year = finfo['year']
    if not ts_type in const.GRID_IO.TS_TYPES:
        raise FileConventionError('Invalid ts_type in file: {}'.format(ts_type))
    elif not -2000 <= year <= 20000:
        raise FileConventionError('Invalid year in file: {}'.format(year))
    try:
        check_time_coord(cube, ts_type, year)
    except UnresolvableTimeDefinitionError as e:
        raise UnresolvableTimeDefinitionError(repr(e))
    except:
        msg = ("Invalid time dimension coordinate in file {}. " 
               .format(os.path.basename(file)))
        logger.warning(msg)
        if const.GRID_IO.CORRECT_TIME_FILENAME:
            logger.warning("Attempting to correct time coordinate "
                           "using information in file name")
            try:
                cube = correct_time_coord(cube, 
                                          ts_type=finfo["ts_type"],
                                          year=finfo["year"]) 
            except:
                pass
        if const.WRITE_FILEIO_ERR_LOG:
            add_file_to_log(file, 'Invalid time dimension')
    return cube
    
def _check_leap_year(num, num_per, ts_type):
    if ts_type == 'daily' and num + 1 == num_per:
        return True
    elif ts_type == '3hourly' and num + 8 == num_per:
        return True
    elif ts_type == 'hourly' and num + 24 == num_per:
        return True
    return False
def check_time_coord(cube, ts_type, year):
    """Method that checks the time coordinate of an iris Cube
    
    This method checks if the time dimension of a cube is accessible and 
    according to the standard (i.e. fully usable). It only checks, and does not
    correct. For the latter, please see :func:`correct_time_coord`.
    
    Parameters
    ----------
    cube : Cube
        cube containing data
    ts_type : str
        pyaerocom ts_type
    year : 
        year of data
    
    Returns
    -------
    bool
        True, if time dimension is ok, False if not
    """
    try:
        t = cube.coord("time")
    except:
        raise AttributeError("Cube does not contain time dimension")
    if not isinstance(t, iris.coords.DimCoord):
        raise AttributeError("Time is not a DimCoord instance")
    try:
        cftime_to_datetime64(0, cfunit=t.units)
    except:
        raise ValueError("Could not convert time unit string")

    tidx = make_datetimeindex(ts_type, year)
    
    num_per = len(tidx)
    num = len(t.points)

    if not num == num_per:
        if tidx[0].is_leap_year:
            if not _check_leap_year(num, num_per, ts_type):
                raise UnresolvableTimeDefinitionError('Expected {} timestamps but '
                                                  'data has {}'
                                                  .format(len(tidx), num))
        else:
            raise UnresolvableTimeDefinitionError('Expected {} timestamps but '
                                                  'data has {}'
                                                  .format(len(tidx), num))
        
    
    pstr = TS_TYPE_TO_PANDAS_FREQ[ts_type]
    # ToDo: check why MS is not working for period conversion
    if pstr == 'MS':
        pstr = 'M'
    # convert first and last timestamps of index array into periods 
    # (e.g. January and December for monthly data)
    per0 = tidx[0].to_period(pstr)
    per1 = tidx[-1].to_period(pstr)    
    
    # first and last timestamp in data
    t0, t1 = cftime_to_datetime64([t.points[0], t.points[-1]], cfunit=t.units)
    
    if not per0.start_time <= t0 <= per0.end_time:
        raise ValueError('First timestamp of data {} does not lie in first '
                         'period: {}'.format(t0, per0))
    elif not per1.start_time <= t1 <= per1.end_time:
        raise ValueError('Last timestamp of data {} does not lie in last '
                         'period: {}'.format(t1, per1))
        

def correct_time_coord(cube, ts_type, year):
    """Method that corrects the time coordinate of an iris Cube
    
    Parameters
    ----------
    cube : Cube
        cube containing data
    ts_type : str
        temporal resolution of data (e.g. "hourly", "daily"). This information
        is e.g. encrypted in the filename of a NetCDF file and may be 
        accessed using :class:`pyaerocom.io.FileConventionRead`
    year : int
        interger specifying year of observation, e.g. 2017
    tindex_cube : int
        index of time dimension in Cube
    
    Returns
    -------
    Cube
        the same instance of the input cube with corrected time dimension axis
        
    """
    tindex_cube = None
    dim_lens = []
    for i, coord in enumerate(cube.dim_coords):
        dim_lens.append(len(coord.points))
        if coord.name() == 'time':
            tindex_cube = i
    if tindex_cube is None:
        if cube.ndim != len(cube.dim_coords): #one dimension is missing
            for idx, dim_len in enumerate(cube.shape):
                if not dim_len in dim_lens: #candidate
                    tindex_cube = idx
    if tindex_cube is None:
        raise NetcdfError('Failed to identify data index of time dimension in '
                          'cube {}'.format(repr(cube)))
    tres_str = TSTR_TO_CF[ts_type]
    conv = TSTR_TO_NP_DT[ts_type]
    tunit_str = '%s since %s-01-01 00:00:00' %(tres_str, year)
    num = cube.shape[tindex_cube]

    tunit = cf_units.Unit(tunit_str, calendar=cf_units.CALENDAR_STANDARD)
    tres_np = TSTR_TO_NP_TD[ts_type]
    base = datetime64("%s-01-01 00:00:00" %year).astype(conv)
    times = base + arange(0, num, 1).astype(tres_np)
    # see this thread https://github.com/matplotlib/matplotlib/issues/2259/
    times_dt = times.astype("datetime64[s]").astype(datetime)
#    timestamps = datetime64(str(year)) + 
    time_nums = [tunit.date2num(t) for t in times_dt]
    tcoord = iris.coords.DimCoord(time_nums, standard_name='time', units=tunit)
        
    #tcoord_dim = cube.coord_dims('time')
    try:
        cube.remove_coord('time')
    except:
        pass
    cube.add_dim_coord(tcoord, tindex_cube)
    return cube

def concatenate_iris_cubes(cubes, error_on_mismatch=True):
    """Concatenate list of :class:`iris.Cube` instances cubes into single Cube
    
    Helper method for concatenating list of cubes and that helps 
    with handling the fact that the corresponding iris method is not well 
    defined in the sense of what it returns (i.e. instance of 
    :class:`Cube` or :class:`CubeList`, depending on whether all cubes 
    could be concatenated or not...)
    
    This method is not supposed to be called directly but rather 
    :func:`concatenate_cubes` (which ALWAYS returns instance of 
    :class:`Cube` or raises Exception) or :func:`concatenate_possible_cubes`
    (which ALWAYS returns instance of :class:`CubeList` or raises Exception)
    
    Parameters
    ----------
    cubes : CubeList
        list of individual cubes
    error_on_mismatch
        boolean specifying whether an Exception is supposed to be raised
        or not
        
    Returns
    -------
    :obj:`Cube` or :obj:`CubeList`
        result of concatenation 
    
    Raises
    ------
    iris.exceptions.ConcatenateError
        if ``error_on_mismatch=True`` and input cubes could not all concatenated
        into a single instance of :class:`iris.Cube` class.
        
    """
    var_name = cubes[0].var_name
    if const.GRID_IO.EQUALISE_METADATA:
        meta_init = cubes[0].metadata
        if not all([x.metadata == meta_init for x in cubes]):
            logger.warning("{} cubes to be concatenated have different meta "
                           "data settings. These will be unified using the "
                           "metadata dictionary of the first cube "
                           "(otherwise the method concatenate of the iris "
                           "package won't work)".format(var_name))
            for cube in cubes:
                cube.metadata = meta_init
                
    #now put the CubeList together and form one cube
    #1st equalise the cubes (remove non common attributes)
    equalise_attributes(cubes)
    #unify time units
    iris.util.unify_time_units(cubes)
    
    #now concatenate the cube list to one cube
    cubes_concat = iris._concatenate.concatenate(cubes, error_on_mismatch)
    
    return cubes_concat[0]

if __name__== "__main__":
    import pyaerocom as pya
    
    r = pya.io.ReadGridded('BCC-CUACE_HIST')
    
    d = r.read_var('zg')
    print(d)
    
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Config file for registered AeroCom observations datasets and variables

UNDER DEVELOPMENT
"""

OBS_SOURCES = {

'EBAS-Lev3'         :   dict(obs_id='EBASMC', 
                             obs_vars=['absc550aer', 'scatc550dryaer'], 
                             vert_scheme='surface',
                             obs_vert_type='Surface'),
        
'AeronetSun'        :   dict(obs_id='AeronetSunV3Lev2.daily',
                             obs_vars=['ang4487aer', 'od550aer'],
                             obs_vert_type='Column'),
                             
'AeronetSDA'        :   dict(obs_id='AeronetSDAV3Lev2.daily',
                             obs_vars=['od550lt1aer', 'od550gt1aer'],
                             obs_vert_type='Column'),

'AeronetInv'        :   dict(obs_id='AeronetInvV3Lev2.daily',
                             obs_vars=['abs550aer'],
                             obs_vert_type='Column'),
'MODIS6-terra'      :   dict(obs_id='MODIS6.terra',
                             obs_vars=['od550aer'],
                             regrid_res_deg=5,
                             obs_vert_type='Column',
                             ts_type='daily',
                             var_outlier_ranges=dict(od550aer=[0,10])),
'MODIS6-aqua'       :   dict(obs_id='MODIS6.aqua',
                             obs_vars=['od550aer'],
                             regrid_res_deg=5,
                             obs_vert_type='Column',
                             ts_type='daily',
                             var_outlier_ranges=dict(od550aer=[0,10])),
'AATSR4.3-SU'       :   dict(obs_id='AATSR_SU_v4.3',
                            obs_vars=['od550aer', 'ang4487aer', 'od550lt1aer',
                                      'od550gt1aer', 'abs550aer'],
                            regrid_res_deg=5,
                            ts_type='daily',
                            obs_vert_type='Column'),
'CALIOPv3'          :   dict(obs_id='CALIOP_V3.00_Cloudfree_day',
                             obs_vars=['od550aer'],
                             regrid_res_deg=5,
                             ts_type='monthly',
                             var_outlier_ranges=dict(od550aer=[0.001,10]),
                             obs_vert_type='Column')
}

OBS_DEFAULTS = {
        
        'Column'        :   {'od550aer'          :  'AeronetSun',
                             'ang4487aer'        :  'AeronetSun',
                             'od550lt1aer'       :  'AeronetSDA',
                             'od550gt1aer'       :  'AeronetSDA',
                             'abs550aer'         :  'AeronetInv'},
                         
        'Surface'       :   {'absc550aer'        :  'EBAS-Lev3',
                             'scatc550dryaer'    :  'EBAS-Lev3'},
}
                
def get_default_obsnetwork(var_name, vert_which=None):
    res = []
    for k, v in OBS_DEFAULTS.items():
        if vert_which is None:
            if var_name in v:
                res.append(v[var_name])
        elif vert_which ==  k and var_name in v:
            return v[var_name]
    num = len(res)
    if num == 0:
        raise ValueError('Could not find default obs network for variable {}'
                         .format(var_name))
    elif num > 1:
        raise ValueError('Found multiple default obs networks for variable {}'
                         'Please specify vertical type via vert_which input arg'
                         .format(var_name))
    return res[0]
        
    
                
def get_all_vars():
    all_vars =  []
    for cfg in OBS_SOURCES.values():
        all_vars.extend(cfg['obs_vars'])
    return list(dict.fromkeys(all_vars))

if __name__ == '__main__':
    
    print(get_all_vars())
    print(get_default_obsnetwork('ang4487aer'))
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug 13 10:49:29 2018

@author: jonasg
"""
import matplotlib.pyplot as plt
import os
import matplotlib.dates as dates

def plot_series_year(obs_data, model_data, var_name, savefig=True,
                     save_dir=None):
    """Plot one year time series overlay of model and station data
    
    Creates default plot of Aerocom time series overlay (see e.g. `here 
    <http://aerocom.met.no/DATA/SURFOBS/TM5_AP3-CTRL2016/plots/
    OD550_AER_an2010_mALLYEAR_ARMDarwin_SERIES_AERONETSun2.0.ps.png>`__ for 
    an example)
    
    Note
    ----
    This is a beta version and thus, might undergo major changes
    
    Todo
    ----
    - Review y axis limits (default variable info)
    - Check date and time formatting
    
    Parameters
    ----------
    obs_data : :obj:`StationData` 
        dictionary containing station meta info and data
    model_data : :obj:`dict`
        dictionary containing model meta info and data at location of obs 
        station
    var_name : str
        variable name
        
    Returns
    -------
    axes
        matplotlib axes instance 
        
    Raises
    ------
    IOError 
        if variable data is not available in input model or obsdata
    """
    if not var_name in model_data:
        raise IOError('Missing {} in model data'.format(var_name))
    elif not var_name in obs_data:
        raise IOError('Missing {} in station data'.format(var_name))
    
    fig = plt.figure(figsize=(12,8))
    ax = fig.add_axes([0.1, 0.1, 0.85, 0.8])
    
    mseries = model_data[var_name]
    oseries = obs_data[var_name]
    idx = mseries.index
    
    #ax.plot()
    
    ax.plot_date(idx.to_pydatetime(), mseries, 'rs--',
                 markersize=8, markerfacecolor='none')
    ax.xaxis.set_minor_locator(dates.MonthLocator(bymonthday=15, 
                                                  interval=1))
    
    ax.xaxis.set_minor_formatter(dates.DateFormatter('%b'))
    ax.xaxis.grid(True, which="minor")
    ax.yaxis.grid()
    ax.xaxis.set_major_locator(dates.YearLocator(month=6, day=15))
    ax.xaxis.set_major_formatter(dates.DateFormatter('\n\n%Y'))
    
    ax.plot(oseries, '-vb', markersize=8, markerfacecolor='none')
    ylim = ax.get_ylim()
    # add some space for printing legend
    ax.set_ylim([0, ylim[1]*1.1])
    ax.set_ylabel(var_name)
    ax.annotate('source: AEROCOM', xy=(0.95, 0.025), 
                xycoords='figure fraction',
                horizontalalignment='right', 
                verticalalignment='center',
                fontsize=10, 
                bbox=dict(boxstyle='square', facecolor='none',
                          edgecolor='black'))
    
    ax.annotate('Obs: {}'.format(obs_data['data_id']),
                xy=(0.97, 0.1),
                xycoords='figure fraction',
                rotation='vertical',
                horizontalalignment='right',
                verticalalignment='bottom',
                fontsize=10)
    
    ax.annotate('PI: {}'.format(obs_data['PI']),
                xy=(0.985, 0.1),
                xycoords='figure fraction',
                rotation='vertical',
                horizontalalignment='right',
                verticalalignment='bottom',
                fontsize=10)
    
    ax.annotate(model_data['name'], xy=(0.105, 0.89),
                xycoords='figure fraction',
                horizontalalignment='left', 
                verticalalignment='top',
                fontsize=16, 
                color='r')
    
    ax.annotate('Obs: {}'.format(obs_data['data_id']),
                xy=(0.945, 0.89),
                xycoords='figure fraction',
                horizontalalignment='right', 
                verticalalignment='top',
                fontsize=16, 
                color='b')
    
    ax.set_title('{} ({:.2f}; {:.2f}; {:.0f}m)'
                 .format(obs_data['station_name'],
                         obs_data['latitude'],
                         obs_data['longitude'],
                         obs_data['altitude']))
    
    if savefig:
        if not os.path.exists(save_dir):
            raise IOError('Ouptut directory {} does not exist'.format(save_dir))
            
        name = '{}_an{}_freq{}_{}_SERIES_{}.png'.format(var_name,
                                                          idx[0].year,
                                                          idx.freqstr,
                                                          obs_data['station_name'],
                                                          obs_data['data_id'])
        fig.savefig(os.path.join(save_dir, name))
        
    return ax


################################################################
# read_aeronet_sunv3.py
#
# read Aeronet direct sun V3 data
#
# this file is part of the pyaerocom package
#
#################################################################
# Created 20180626 by Jan Griesfeller for Met Norway
#
# Last changed: See git log
#################################################################

# Copyright (C) 2018 met.no
# Contact information:
# Norwegian Meteorological Institute
# Box 43 Blindern
# 0313 OSLO
# NORWAY
# E-mail: jan.griesfeller@met.no
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA
import os

import numpy as np
import pandas as pd

from pyaerocom import const
from pyaerocom.mathutils import (calc_ang4487aer, calc_od550aer)
from pyaerocom.io.readaeronetbase import ReadAeronetBase
from pyaerocom.stationdata import StationData

class ReadAeronetSunV3(ReadAeronetBase):
    """Interface for reading Aeronet direct sun version 3 Level 1.5 and 2.0 data

    .. seealso::
        
        Base classes :class:`ReadAeronetBase` and :class:`ReadUngriddedBase`

    """
    #: Mask for identifying datafiles 
    _FILEMASK = '*.lev30'
    
    #: version log of this class (for caching)
    __version__ = '0.08_' + ReadAeronetBase.__baseversion__
    
    #: Name of dataset (OBS_ID)
    DATA_ID = const.AERONET_SUN_V3L2_AOD_DAILY_NAME
    
    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.AERONET_SUN_V3L15_AOD_DAILY_NAME,
                          const.AERONET_SUN_V3L15_AOD_ALL_POINTS_NAME,
                          const.AERONET_SUN_V3L2_AOD_DAILY_NAME,
                          const.AERONET_SUN_V3L2_AOD_ALL_POINTS_NAME]
    
    #: dictionary assigning temporal resolution flags for supported datasets
    #: that are provided in a defined temporal resolution
    TS_TYPES = {const.AERONET_SUN_V3L15_AOD_DAILY_NAME :    'daily',
                const.AERONET_SUN_V3L2_AOD_DAILY_NAME  :    'daily'}
    
    #: default variables for read method
    DEFAULT_VARS = ['od550aer', 'ang4487aer']
    
    #: value corresponding to invalid measurement
    #NAN_VAL = -9999.
    NAN_VAL = -999.
    
    #: Mappings for identifying variables in file
    VAR_PATTERNS_FILE = {'AOD_([0-9]*)nm' : 'od*aer'}
    
    #: dictionary specifying the file column names (values) for each Aerocom 
    #: variable (keys)
    VAR_NAMES_FILE = {}
    VAR_NAMES_FILE['od340aer'] = 'AOD_340nm'
    VAR_NAMES_FILE['od440aer'] = 'AOD_440nm'
    VAR_NAMES_FILE['od500aer'] = 'AOD_500nm'
    # VAR_NAMES_FILE['od865aer'] = 'AOD_865nm'
    VAR_NAMES_FILE['od870aer'] = 'AOD_870nm'
    VAR_NAMES_FILE['ang4487aer'] = '440-870_Angstrom_Exponent'

    #: dictionary specifying the file column names (values) for each 
    #: metadata key (cf. attributes of :class:`StationData`, e.g.
    #: 'station_name', 'longitude', 'latitude', 'altitude')
    META_NAMES_FILE = {}
    META_NAMES_FILE['data_quality_level'] = 'Data_Quality_Level'
    META_NAMES_FILE['instrument_number'] = 'AERONET_Instrument_Number'
    META_NAMES_FILE['station_name'] = 'AERONET_Site'
    META_NAMES_FILE['latitude'] = 'Site_Latitude(Degrees)'
    META_NAMES_FILE['longitude'] = 'Site_Longitude(Degrees)'
    META_NAMES_FILE['altitude'] = 'Site_Elevation(m)'
    META_NAMES_FILE['date'] = 'Date(dd:mm:yyyy)'
    META_NAMES_FILE['time'] = 'Time(hh:mm:ss)'
    META_NAMES_FILE['day_of_year'] = 'Day_of_Year'
    
    #: dictionary containing information about additionally required variables
    #: for each auxiliary variable (i.e. each variable that is not provided
    #: by the original data but computed on import)
    AUX_REQUIRES = {'ang4487aer_calc'   :   ['od440aer',
                                             'od870aer'],
                    'od550aer'          :   ['od440aer', 
                                             'od500aer',
                                             'ang4487aer']}
                    
    #: Functions that are used to compute additional variables (i.e. one 
    #: for each variable defined in AUX_REQUIRES)
    AUX_FUNS = {'ang4487aer_calc'   :   calc_ang4487aer,
                'od550aer'          :   calc_od550aer}
    
    #: List of variables that are provided by this dataset (will be extended 
    #: by auxiliary variables on class init, for details see __init__ method of
    #: base class ReadUngriddedBase)
    PROVIDES_VARIABLES = list(VAR_NAMES_FILE.keys())

    def read_file(self, filename, vars_to_retrieve=None, 
                  vars_as_series=False, read_all_possible=False):
        """Read Aeronet Sun V3 level 1.5 or 2 file 

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        vars_as_series : bool
            if True, the data columns of all variables in the result dictionary
            are converted into pandas Series objects
        read_all_possible : bool
            if True, than all available variables belonging to either of the 
            variable families that are specified in :attr:`VAR_PATTERNS_FILE` 
            are read from the file (in addition to the ones that are specified
            via vars_to_retrieve).  
            
        Returns
        -------
        StationData
            dict-like object containing results
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        # implemented in base class
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)
       
        #create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out.data_id = self.DATA_ID
        # create empty arrays for meta information
        for item in self.META_NAMES_FILE:
            data_out[item] = []
        
        # Iterate over the lines of the file
        self.logger.info("Reading file {}".format(filename))
        with open(filename, 'rt') as in_file:
            _lines_ignored = []
            _lines_ignored.append(in_file.readline())
            _lines_ignored.append(in_file.readline())
            _lines_ignored.append(in_file.readline())
            _lines_ignored.append(in_file.readline())
            # PI line
            dummy_arr = in_file.readline().strip().split(';')
            data_out['PI'] = dummy_arr[0].split('=')[1]
            data_out['PI_email'] = dummy_arr[1].split('=')[1]
            data_out['ts_type'] = self.TS_TYPE

            data_type_comment = in_file.readline()
            _lines_ignored.append(data_type_comment)
            # TODO: delete later
            self.logger.debug("Data type comment: {}".format(data_type_comment))
            
            # put together a dict with the header string as key and the index number as value so that we can access
            # the index number via the header string
            col_index_str = in_file.readline()
            if col_index_str != self._last_col_index_str:
                self.logger.info("Header has changed, reloading col_index map")
                self._update_col_index(col_index_str, 
                                       use_all_possible=read_all_possible)
            col_index = self.col_index
            
            # create empty arrays for all variables that are supposed to be read
            # from file
            for var in vars_to_read:
                data_out[var] = []
            # dependent on the station, some of the required input variables
            # may not be provided in the data file. These will be ignored
            # in the following list that iterates over all data rows and will
            # be filled below, with vectors containing NaNs after the file 
            # reading loop
            vars_available = {}
            for var in vars_to_read:
                if var in col_index:
                    vars_available[var] = col_index[var]
                else:
                    self.logger.warning("Variable {} not available in file {}"
                                        .format(var, os.path.basename(filename)))

            for line in in_file:
                # process line
                dummy_arr = line.split(self.COL_DELIM)
                
                # copy the meta data (array of type string)
                for var in self.META_NAMES_FILE:
                    val = dummy_arr[col_index[var]]
                    try:
                        # e.g. lon, lat, altitude
                        val = float(val)
                    except:
                        pass
                    data_out[var].append(val)
                
                # This uses the numpy datestring64 functions that e.g. also 
                # support Months as a time step for timedelta
                # Build a proper ISO 8601 UTC date string
                day, month, year = dummy_arr[col_index['date']].split(':')
                datestring = '-'.join([year, month, day])
                datestring = 'T'.join([datestring, dummy_arr[col_index['time']]])
                # NOTE JGLISS: parsing timezone offset was removed on 22/2/19
                # since it is deprecated in recent numpy versions, for details
                # see https://www.numpy.org/devdocs/reference/arrays.datetime.html#changes-with-numpy-1-11
                # datestring = '+'.join([datestring, '00:00'])
                
                data_out['dtime'].append(np.datetime64(datestring))

                # TODO: remove elif if ensured that it works
                for var, idx in vars_available.items():
                    val = np.float_(dummy_arr[idx])
                    if val == self.NAN_VAL: 
                        val = np.nan
                    elif val < self.NAN_VAL:
                        raise Exception('Developers: please debug')
                    data_out[var].append(val)

        # convert all lists to numpy arrays
        data_out['dtime'] = np.asarray(data_out['dtime'])
        
        for item in self.META_NAMES_FILE:
            data_out[item] = np.asarray(data_out[item])
            
        for var in vars_to_read:
            if var in vars_available:
                array = np.asarray(data_out[var])
            else:
                array = np.zeros(len(data_out['dtime'])) * np.nan
            data_out[var] = array
        
        # compute additional variables (if applicable)
        data_out = self.compute_additional_vars(data_out, vars_to_compute)
        
        # convert data vectors to pandas.Series (if applicable)
        if vars_as_series:        
            for var in (vars_to_read + vars_to_compute):
                if var in vars_to_retrieve:
                    data_out[var] = pd.Series(data_out[var], 
                                              index=data_out['dtime'])
                else:
                    del data_out[var]
        self.logger.debug('The following lines were ignored: {}'.format(
                          _lines_ignored))
        return data_out

if __name__=="__main__":
    import matplotlib.pyplot as plt
    plt.close('all')
    
    read = ReadAeronetSunV3()
    files = read.get_file_list(pattern='*Berlin*')
    data = read.read_file(files[0])#, read_all_possible=True)
    #data = read.read(['od550aer', 'ang4487aer'])
    
    #stat = data.to_station_data('Palgrunden')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
ToDos
-----

- Review outlier removal
"""
from collections import OrderedDict as od
import datetime
from fnmatch import fnmatch
import numpy as np
import os, glob
import pandas as pd
from pyaerocom import const, __version__, Region
from pyaerocom._lowlevel_helpers import (check_dirs_exist, dict_to_str)
from pyaerocom.helpers import isnumeric
from pyaerocom.trends_helpers import (_init_trends_result_dict, 
                                      _compute_trend_error,
                                      _get_season, _mid_season, 
                                      _find_area, _years_from_periodstr)
from pyaerocom.io.helpers import save_dict_json
from pyaerocom.io import ReadGridded
from pyaerocom.web.var_groups import var_groups
from pyaerocom.web.helpers import (ObsConfigEval, ModelConfigEval,
                                   update_menu_trends_iface,
                                   get_all_config_files_trends_iface)
from pyaerocom.exceptions import DataCoverageError, MetaDataError
from scipy.stats import kendalltau
from scipy.stats.mstats import theilslopes
import simplejson

class TrendsEvaluation(object):
    """High-level analysis class to compute json files for trends interface
    
    Web: https://aerocom-trends.met.no/
    
    Attributes
    ----------
    
    name : str
        name of this configuration setup
    config_dir : str, optional
        directory containing available configuration files
    out_basedir : str
        basic output directory
    out_dirs : dict
        dictionary containing output directories relative to `out_basedir`
        (will be filled automatically in :func:`init_dirs`)
    obs_config : dict
        dictionary containing configuration details for individual observations
        (i.e. instances of :class:`ObsConfigEval` for each observation) used 
        for the analysis.
    obs_ignore : list, optional
        list of observations that are supposed to be ignored in analysis 
        (keys from :attr:`obs_config`)
    model_config : dict
        dictionary specifying models or satellites to be used for colocation
        with the observations.
    periods : list
        list of periods used for the trends analysis
    regions : dict
        dictionary containing regions used for regional statistics. Keys 
        are region names as displayed in web interface and values are 
        instances of :class:`pyaerocom.Region` for each of the regions.
    var_mapping : dict
        mapping of variable names for menu in interface
    var_order_menu : list, optional
        order of variables in menu.
    obs_order_menu_cfg : bool
        where applicable and if True, then order observations with common 
        variables in the menu in the same order as they appear in 
        :attr:`obs_config`. If False, order alphabetically.
    slope_alpha : float
        desired confidence of trends slope retrieval (`alpha` parameter 
        in :func:`scipy.stats.theilslopes`), e.g. .68 corresponds to 1sigma
        confidence (68%).
    min_dim : int
        minimum number of measurements per month in order to compute monthly
        averages (does not apply for input data that is already aggregated to
        monthly).
    avg_how : str
        averaging method for temporal aggregation (see :attr:`AVG_OPTS` for 
        allowed values).
    clear_existing_json : bool
        If true, then existing json files for a given run will be deleted 
        before rerunning (per entry in :attr:`obs_config`).
    write_logfiles : bool
        if True, then logfiles will be written for outliers that were 
        identified, errors that occurred and files that were created in the 
        analysis (per entry in :attr:`obs_config`).
    logdir : str
        directory where logfiles are stored.
        
    Parameters
    ----------
    config_file : str, optional
        if provided and valid, then configuration is loaded from this file
    **settings
        configuration settings
    """
    #: Vertical layer ranges
    VERT_LAYERS = {'0-2km'  :   [0, 2000],
                   '2-5km'  :   [2000, 5000],
                   '5-10km' :   [5000, 10000]}
    
    #: Names of output directories relative to :attr:`out_basedir`
    OUT_DIR_NAMES = ['ts', 'map']
    
    #: available averaging options for timeseries
    AVG_OPTS = ['mean', 'median']
    
    #: mapping of metadata names between pyaerocom (keys) and json trends files
    #: (values)
    KEYMAP = od(var_name        = 'var_name',
                station_name    = 'station',
                latitude        = 'lat',
                longitude       = 'lon',
                altitude        = 'alt',
                data_id         = 'data_id',
                dataset_name    = 'dataset',
                data_product    = 'product',
                data_version    = 'data_version',
                data_level      = 'data_level',
                website         = 'website',
                instrument_name = 'instrument',
                ts_type_src     = 'freq_src',
                ts_type         = 'freq',
                PI              = 'PI',
                wavelength      = 'wavelength',
                units           = 'unit',
                statistics      = 'statistics',
                matrix          = 'size_info',
                overlap_info    = 'data_overlap')
    
    #: supported frequency codes
    TS_TYPES = const.GRID_IO.TS_TYPES
    
    #: Definition of regions for trends interface
    # ToDo: use pyaerocom regions
    DEFAULT_REGIONS = {
            'NAMERICA': 'N-America',
            'SAMERICA': 'S-America',
            'EUROPE': 'Europe',
            'ASIA': 'Asia',
            #'CHINA': 'China',
            'AUSTRALIA': 'Australia',
            'NAFRICA': 'N-Africa',
            'SAFRICA': 'S-Africa',
            #'INDIA': 'India',
            'WORLD': 'World'
    }
    SEASONS = ['spring','summer','autumn','winter','all']
    
    JSON_CFG_IGNORE = ['regions', '_log', 'out_dirs']
    
    GRIDDED_TS_TYPE = 'monthly'
    def __init__(self, config_file=None, **settings):
        
        self.name = None
        self._log = const.print_log
        
        self.config_dir = None
        self.out_basedir = None
        self.out_dirs = {}
        
        self.obs_config = {}
        self.obs_ignore = []
        
        self.model_config = {}
        
        self.periods = []
        self.regions = {}
        self._add_regions = {}
        
        self.var_mapping = {}
        self.var_order_menu = []
        
        self.obs_order_menu_cfg = True
        
        # options
        self.slope_alpha = 0.68 
        self.min_dim = 5
        self.avg_how = 'mean'
        
        self.delete_outdated = True
        self.clear_existing_json = True
        self.write_logfiles = True
        self.logdir = None
        
        # initialise
        if config_file is not None: # user provided configuration file
            self.from_json(config_file)
        else: # user maybe provided full config via dict, or only name
            self.update(**settings)
    
    def __setitem__(self, key, val):
        
        if key == 'obs_config':
            self._set_obsconfig(val)
        elif key == 'model_config':
            self._set_modelconfig(val)
        elif isinstance(key, str) and isinstance(val, dict):
            if 'obs_id' in val:
                if key in self.obs_config:
                    self._log.warn('Obs config for key {}  already exists and '
                                   'will be overwritten {}'.format(key))
                self.obs_config[key] = ObsConfigEval(**val)
            elif 'model_id' in val:
                if not 'mtype' in val:
                    raise KeyError('Need key "mtype" in specfication of model {}'
                                   .format(key))
                if key in self.model_config:
                    self._log.warn('Model config for key {}  already exists and '
                                   'will be overwritten {}'.format(key))
                self.model_config[key] = ModelConfigEval(**val)
            else:
                self.__dict__[key] = val
        elif key in self.__dict__:
            self.__dict__[key] = val
        else:
            raise KeyError('Invalid input key {}. Cannot assign {}'
                           .format(key, val))
                
    def __getitem__(self, key):
        if not key in self.__dict__:
            raise KeyError('No such key available: {}'.format(key))
        return self.__dict__[key]
        
    def __str__(self):
        indent = 2
        _indent_str = indent*' '
        head = "Pyaerocom {}".format(type(self).__name__)
        s = "\n{}\n{}".format(head, len(head)*"-")
        obs_cfg = ''
        for key, val in self.__dict__.items():
            if key[0] == '_':
                continue
            elif key == 'obs_config':
                for k, v in val.items():
                    obs_cfg += '\n\n{}{}:'.format(_indent_str, k)
                    obs_cfg = dict_to_str(v, obs_cfg, 
                                          indent=indent+2)
            elif isinstance(val, dict):
                s += '\n{}:'.format(key)
                s = dict_to_str(val, s)
            else:
                s += '\n{}: {}'.format(key, val)
        s += obs_cfg
        return s
    
    
    def get_config_files(self, config_dir):
        """Get all available configuration files
        
        Parameters
        ----------
        config_dir : str
            directory containing configuration files
        
        Returns
        -------
        dict
            dictionary containing all available configuration files
        
        Raises
        ------
        FileNotFoundError
            if no config files could be found in input directory
        """
        files = get_all_config_files_trends_iface(config_dir)
        if len(files) > 0:
            return files
        raise FileNotFoundError('Could not find any configuration files')
            
    @property
    def menu_file(self):
        """json file containing region specifications"""
        return os.path.join(self.out_basedir, 'menu.json')
    
    @property
    def all_obs_names(self):
        """List of all obs names"""
        return list(self.obs_config)
    
    @property
    def all_map_files(self):
        """List of all existing map files"""
        return sorted([f for f in os.listdir(self.out_dirs['map']) 
                       if f.endswith('.json') and f.startswith('OBS-')])
    
    def check_config(self):
        """Check if there are any problems in the configuration
        
        Returns
        -------
        bool
            True if all is good, False, if configuration settings are missing
        
        Raises
        ------
        ValueError
            if one of the provided period strings is invalid
        AttributeError
            if model configuration is missing
        AssertionError
            if there is a problem with EBAS default configuration
        """
        if len(self.periods) == 0:
            return False
        elif len(self.obs_config) == 0:
            return False
        for p in self.periods:
            try:
                start, stop = _years_from_periodstr(p)
            except:
                raise ValueError('Invalid input for period: {}. '
                                 'Need string in format "2010-2019"'.format(p))
        
        for run, cfg in self.obs_config.items():
            if '_' in run:
                raise AttributeError('Invalid name: {} (no underscores allowed)'
                                     .format(run))
            if cfg.obs_id == 'EBASMC':
                check_ebas_default()
            if 'models' in cfg:
                for model in cfg['models']:
                    if not model in self.model_config:
                        raise AttributeError('No such model with name {} '
                                             'available in model_config...'
                                             .format(model))
        return True
        
        
    
    def find_obs_matches(self, name_or_pattern):
        """Find model names that match input search pattern(s)
        
        Parameters
        ----------
        name_or_pattern : :obj:`str`, or :obj:`list`
            Name or pattern specifying obs search string. Can also be a list 
            of names or patterns to search for multiple obs networks.
            
        Returns
        -------
        list
            list of model names (i.e. keys of :attr:`obs_config`) that match 
            the input search string(s) or pattern(s)
            
        Raises
        ------
        KeyError
            if no matches can be found
        """
        
        
        if isinstance(name_or_pattern, str):
            name_or_pattern = [name_or_pattern]
        matches = []
        for search_pattern in name_or_pattern:
            for mname in self.obs_config:
                if fnmatch(mname, search_pattern) and not mname in matches:
                    matches.append(mname)
        if len(matches) == 0:
            raise KeyError('No models could be found that match input {}'
                           .format(name_or_pattern))
        return matches
    
    def check_model_access(self, models, vars_to_retrieve):
        """Check availability of input variables in input model IDs
        
        Parameters
        ----------
        models : list
            list of model names that available in :attr:`model_config`.
        vars_to_retrieve : list
            list of variable names that are supposed to be checked in models
            
        Returns
        -------
        dict
            keys are models and values are lists of variables that are 
            available in each of these models
        """
        maccess = {}
        for model_name in models:
            if not model_name in self.model_config:
                raise AttributeError('No model {} available in model_config'
                                     .format(model_name))
            mcfg = self.model_config[model_name]
            model_id = mcfg['model_id']
            try:
                r = ReadGridded(model_id)
                vars_avail = {}
                for obs_var in vars_to_retrieve:
                    try:
                        mod_var = mcfg['model_use_vars'][obs_var]
                    except KeyError:
                        mod_var = obs_var
                    if mod_var in r.vars_provided:
                        vars_avail[obs_var] = mod_var
                if len(vars_avail) > 0:
                    maccess[model_name] = dict(model_id=model_id,
                                               vars_avail=vars_avail)
            except:
                self._log.warning('Model {} does not provide any of the req. '
                                  'variables {}'.format(model_name, 
                                                        vars_to_retrieve))
            
                    
        return maccess
    
    def load_ungridded(self, network_id, vars_to_retrieve, **constraints):
        """Load obsdata into instance of :class:`pyaerocom.UngriddedData`
        
        Parameters
        ----------
        network_id : str
            Name of network (if you don't know the options, check 
            `pyaerocom.const.OBS_IDS`)
        vars_to_retrieve 
            str or list / tuple of strings specifying AEROCOM variables that are 
            supposed to be imported
        **constraints
            Further reading constraints passed to :func:`UngriddedData.read` and
            from there, further passed to :func:`read`  of respective network
            reading class. 
            
        Returns
        -------
        UngriddedData
        """
        import pyaerocom as pya
        r = pya.io.ReadUngridded()
        return r.read(datasets_to_read=network_id,
                      vars_to_retrieve=vars_to_retrieve,
                      **constraints)
        
    def run_evaluation(self, obs_name=None, update_menu=True, 
                       write_logfiles=None, clear_existing_json=None):
        """Run trends evaluation (main function of this interface)
        
        Note
        ----
        Individual runs are / can be specified in :attr:`obs_config`
        
        Parameters
        -----------
        obs_name : str or list, optional
            name(s) of run(s). If None, then all runs available in 
            :attr:`obs_config` are used.
        update_menu : bool
            if True, the menu.json file is automatically updated (if 
            applicable)
        write_logfiles : bool, optional
            option to write logfiles (uses :attr:`write_logfiles` if None,
            else, updates the latter)
        clear_existing_json : bool, optional
            option to write logfiles (uses :attr:`clear_existing_json` if None,
            else, updates the latter)
        
        Returns 
        -------
        dict
            nested dictionary containing information about the individual runs 
            (keys), that is, the output from :func:`run_single` for each of the 
            runs.
        """
        if self.delete_outdated:
            self.delete_outdated_json_files()
            
        if write_logfiles is not None:
            self.write_logfiles = write_logfiles
        if clear_existing_json is not None:
            self.clear_existing_json = clear_existing_json
        if obs_name is None:
            obs_list = list(self.obs_config)
        else:
            obs_list = self.find_obs_matches(obs_name)
        result = {}
        
        for obs_name in obs_list:
            result[obs_name] = self.run_single(obs_name, write_logfiles)
        for name, files_written in result.items():
            print('Files written:')
            print('RUN: {}'.format(name))
            for k, v in files_written.items():
                print('Folder: {}'.format(k))
                print('Files: {}'.format(v))
        
        self.make_regions_json()
        self.to_json()
        if update_menu:
            self.update_menu()
            
        return result
    
    def get_var_groups(self, obs_name):
        """Split list of variables of obs into variable sublists 
        
        Splitting is done based on predefined var groups, defined in 
        :mod:`var_groups`, where available.
        
        Parameters
        ----------
        obs_id : str
            Name of obsdata source (key of :attr:`obs_config`)
        
        Returns
        -------
        list
            list containing sublists for different variable groups
        """
        obs_id = self.obs_config[obs_name]['obs_id']
        vars_to_retrieve = self.obs_config[obs_name]['obs_vars']
        if not obs_id in var_groups:
            return [vars_to_retrieve]
        result = []
        groups =  var_groups[obs_id]
        added = []
        for group in groups:
            subset = list(np.intersect1d(vars_to_retrieve, group))
            if any([x in added for x in subset]):
                raise ValueError('Fatal: one variable appears to exist in 2 groups')
            added.extend(subset)
            if len(subset) > 0:
                result.append(subset)
        rest = [x for x in vars_to_retrieve if not x in added]
        if len(rest) > 0:
            result.append(rest)
        return result
            
    def run_single(self, obs_name, write_logfiles):
        """Run one of the config entries in :attr:`obs_config`
        
        Parameters
        ----------
        obs_name : str
            name of run (key in :attr:`obs_config`)
        write_logfiles : bool
            write logfiles
        
        Returns
        -------
        dict
            map and ts files that were created
        """
        
        if not obs_name in self.obs_config:
            raise ValueError('No such run with name {}'.format(obs_name))
        config = self.obs_config[obs_name]
        if self.clear_existing_json:
            print('Deleting existing files for run {}'.format(obs_name))
            self.clear_existing(obs_name)
            
        
        obs_id = config['obs_id']
        
        self._log.info('Running {} (NETWORK {})'.format(obs_name, obs_id))
            
        constraints = config['read_opts_ungridded']
        if constraints is None:
            constraints = {}
        
        outlier_log, err_log, files_log = None, None, None
        if write_logfiles:
            outlier_log, err_log, files_log = self._init_logfiles(obs_name)
            
        if 'var_outlier_ranges' in config:
            min_max = config['var_outlier_ranges']
        else:
            min_max = {}
        if 'min_dim' in config:
            min_dim = config['min_dim']
        else:
            min_dim = self.min_dim
        if 'models' in config:
            models = config['models']
        else:
            models = None
            
        files_created = {'ts'  : [],
                         'map' : []}
        
        var_lists = self.get_var_groups(obs_name)
        for vars_to_retrieve in var_lists:
            ungridded_data = self.load_ungridded(obs_id, vars_to_retrieve, 
                                                 **constraints)
            
            if 'obs_filters' in config:
                filters=config['obs_filters']
                ungridded_data = ungridded_data.apply_filters(var_outlier_ranges=min_max,
                                                              **filters)
            
            if config['obs_vert_type']=='Profile':
                files_created = self._run_single_3d(ungridded_data=ungridded_data, 
                                                    vars_to_retrieve=vars_to_retrieve,
                                                    min_dim=min_dim,
                                                    models=models, 
                                                    name=obs_name, 
                                                    files_created=files_created, 
                                                    err_log=err_log)
            else:
                files_created = self._run_single_2d(ungridded_data=ungridded_data, 
                                                    vars_to_retrieve=vars_to_retrieve,
                                                    min_dim=min_dim,
                                                    models=models, 
                                                    name=obs_name,
                                                    vert_which=config['obs_vert_type'],
                                                    files_created=files_created, 
                                                    err_log=err_log)
        
        if write_logfiles:
            files_log.write('MAP files\n---------------------------\n')
            for f in files_created['map']:
                files_log.write('{}\n'.format(f))
            files_log.write('Station ts files\n---------------------------\n')
            for f in files_created['ts']:
                files_log.write('{}\n'.format(f))
                
        if write_logfiles:
            outlier_log.close()
            err_log.close()
            files_log.close()
        return files_created
    
    
        
    def update_menu(self):
        """Update menu.json based on available runs"""
        update_menu_trends_iface(self)
    
    @property
    def regions_file(self):
        """File path of regions.json file"""
        return os.path.join(self.out_basedir, 'regions.json')
    
    def make_regions_json(self):
        """Make regions.json file for web interface"""
        regs = {}
        for regname, reg in self.regions.items():
            
            regs[regname] = r = {}
            
            latr = reg.lat_range
            lonr = reg.lon_range
            r['minLat'] = latr[0]
            r['maxLat'] = latr[1]
            r['minLon'] = lonr[0]
            r['maxLon'] = lonr[1]
        save_dict_json(regs, self.regions_file)
        return regs
    
    def update(self, **settings):
        """Update current setup"""
        for k, v in settings.items():
            self[k] = v
        if self.out_basedir is not None:
            self._init_dirs()
        self._init_regions()
        if not self.check_config():
            self._log.info('Setup is incomplete')
            
    def search_config_file(self, autoload=True):
        """Try to find valid configuration file"""
        if self.name is None:
            raise AttributeError('Attr. name needs to be specified')
        name = self.name
        file = None
        if self.config_dir is not None:
            try:
                files = self.get_config_files(self.config_dir)
                if not name in files:
                    raise FileNotFoundError('No config {} in {}'.format(name, 
                                            self.config_dir))
                file = files[name]
            except:
                pass
        else:
            try:
                files = self.get_config_files('.')
                if not name in files:
                    raise FileNotFoundError('No config {} in current '
                                            'directory'.format(name))
                file = files[name]
            except:
                pass
        if file is None:
            raise FileNotFoundError('Could not find a valid configuration file')
        if autoload:
            self._log.info('Importing configuration from {}'.format(file))
            self.from_json(file)
        return file
                
    def from_json(self, config_file):
        """Load configuration from json config file"""
        with open(config_file, 'r') as f:
            current = simplejson.load(f)
        self.update(**current)  
    
    def to_dict(self):
        """Convert configuration to dictionary"""
        d = {}
        for key, val in self.__dict__.items():
            if key in self.JSON_CFG_IGNORE:
                continue
            elif isinstance(val, dict):
                if key == 'model_config':
                    sub = self._model_config_asdict()
                elif key == 'obs_config':
                    sub = self._obs_config_asdict()
                else:
                    sub = {}
                    for k, v in val.items():
                        if v is not None:
                            sub[k] = v
                d[key] = sub
            else:
                d[key] = val
        return d    
    
    def to_json(self, output_dir=None, filename=None):
        """Convert configuration to json ini file"""
        if filename is None:
            if self.name is None:
                raise ValueError('Attr. "name" needs to be specified')
            filename = 'cfg_{}.json'.format(self.name)
        if not filename.startswith('cfg_'):
            raise ValueError('json configuration file MUST start with cfg_')
        elif not filename.endswith('.json'):
            filename.split('.')[0] += '.json'
        if output_dir is None:
            output_dir = self.out_basedir
        d = self.to_dict()
        
        fp = os.path.join(output_dir, filename)
        save_dict_json(d, fp, indent=3)
        return fp
        
    def print_run_ids(self, check_existing=True, detailed=False):
        """Prints all available run ID's 
        
        Parameters
        ----------
        check_existing : bool
            if True, a check is performed for each run ID, whether .json files 
            exist in the data repository
        detailed : bool
            if True, the corresponding settings for the run are printed too
        """
        for run, info in self.obs_config.items():
            print('Run ID: {}'.format(run))
            if detailed:
                for item, value in info.items():
                    print('{}: {}'.format(item, value))
    
            if check_existing:
                there = False
                try:
                    files = os.listdir(self.out_dirs['map'])
                    for file in files:
                        if run in file:
                            there = True
                            break
                    print('data files (json) exist: {}'.format(there))
                except:
                    print('Failed to check for existing runs for ID {}'.format(run))
            print()
            
    def _get_meta_from_json(self, f):
        """Extract meta information from json file
        
        Parameters
        ----------
        f : str
            map or ts json file
        
        Returns
        -------
        str
            obs_name
        str
            obs_var
        str
            vert_code
        str
            mod_name
        str
            mod_var
        """
        
        obs_info = f.split('OBS-')[1].split('_MOD')[0].split('_')
        obs_name, obs_var = obs_info[0].split(':')
        vert_code = obs_info[1]
        mod_name, mod_var =  f.split('MOD-')[1].split('.json')[0].split(':')
        
        return (obs_name, obs_var, vert_code, mod_name, mod_var)
    
    def get_web_overview_table(self):
        """Computes overview table based on existing map files
        
        Note
        ----
        Ignores map files that belong to runs which are not specified in this
        configuration.
        """
        tab = []
        header = ['obs_name' , 'obs_id', 'obs_var', 
                  'vert', 'mod_name', 'mod_id', 'mod_var', 'mod_type',
                  'periods']
        periods = self.get_periods_from_map_files()
        for f in self.all_map_files:
            (obs_name, obs_var, 
             vert_code, 
             mod_name, mod_var) = self._get_meta_from_json(f)
                
            if not obs_name in self.obs_config:
                continue
            
            obs_id = self.obs_config[obs_name]['obs_id']
            
            mod_id, mod_type = None, None
            if mod_name != 'None':
                mod_id = self.model_config[mod_name]['model_id']
                mod_type = self.model_config[mod_name]['mtype']
            else:
                mod_var = None
                
            #os.path.exists(t.out_dirs['map'] + '/' + t.all_map_files[0])
            tab.append([obs_name, obs_id, obs_var, vert_code, 
                        mod_name, mod_id, mod_var, mod_type, periods[f]])
            
        return pd.DataFrame(tab, columns=header)
    
    def delete_outdated_json_files(self):
        """Delete all json files that are not part of this configuration"""
        for f in self.all_map_files:
            (obs_name, obs_var, 
             vert_code, 
             mod_name, mod_var) = self._get_meta_from_json(f)
                
            if not obs_name in self.obs_config:
                ts_files = glob.glob('{}/OBS-{}:{}_{}_MOD-{}:{}*.json'
                                     .format(self.out_dirs['ts'],
                                             obs_name, obs_var, vert_code,
                                             mod_name, mod_var))
                self._log.info('REMOVING {} files in ts directory for {} ({})'
                               .format(len(ts_files), obs_name, obs_var))
                for file in ts_files:
                    os.remove(file)
                
                self._log.info('REMOVING map file {}'.format(f))
                os.remove(os.path.join(self.out_dirs['map'], f))
        
    def get_periods_from_map_files(self):
        """Get data periods covered for each of the map files
        
        Returns
        -------
        dict
            keys are filenames of map files, values are list of periods 
            covered
        """
        res = {}
        for f in self.all_map_files:
            with open(os.path.join(self.out_dirs['map'], f), 'r') as fp:
                current = simplejson.load(fp)
                res[f] = list(current[0]['all'].keys())
        return res
    
    def get_obsvar_name_and_type(self, obs_var):
        """Get menu name and type of observation variable
        
        Parameters
        ----------
        obs_var : str
            Name of observation variable
        
        Returns
        -------
        str
            menu name of this variable
        str
            menu category of this variable
        """
        try:
            name, tp, cat = self.var_mapping[obs_var]
        except:
            name, tp, cat = obs_var, 'UNDEFINED', 'UNDEFINED'
            self._log.warning('Missing menu name definition for var {}. '
                              'Using variable name'.format(obs_var))
        return (name, tp, cat)
    
    def clear_existing(self, obs_name):
        """Delete existing json files for a certain run
        
        Note
        ----
        Clearing old files is recommended before a new run is started 
        for this ID.
        
        Parameters
        ----------
        obs_name : str
            name of observation (key of :attr:`obs_config`)
        """
        
        for d in self.out_dirs.values():
            for f in os.listdir(d):
                if f.split('_')[0] == obs_name:
                    self._log.info('Removing file {}'.format(f))
                    os.remove(os.path.join(d, f))
                    
    def _init_regions(self):
        """Initiate regions to assign lat / lon coordinates"""
        regs = {}
        for reg_id, reg_name in self.DEFAULT_REGIONS.items():
            regs[reg_name] = Region(reg_id)
        for reg_name, info in self._add_regions.items():
            try:
                regs[reg_name] = Region(name=reg_name, 
                                        lat_range=info['lat_range'],
                                        lon_range=info['lon_range'])
            except:
                self._log.warning('Failed to add region {}: {}'
                                  .format(reg_name, info))
        self.regions = regs
           
    def _init_dirs(self, out_basedir=None):
        """Check and create directories"""
        if out_basedir is not None:
            self.out_basedir = out_basedir
        if self.out_basedir is None:
            self.out_basedir = const.OUTPUTDIR
        check_dirs_exist(self.out_basedir)
        outdirs = {}
        for dname in self.OUT_DIR_NAMES:
            outdirs[dname] = os.path.join(self.out_basedir, dname)
        check_dirs_exist(**outdirs)
        self.out_dirs = outdirs
        
        if self.logdir is None:
            self.logdir = os.path.join(const.OUTPUTDIR, 'logfiles')
        check_dirs_exist(self.logdir)
        self._check_output_dirs()
    
    def _check_output_dirs(self):
        """Check if relevant output directories for json files exist"""
        dirs = self.out_dirs
        for k, d in dirs.items():
            if not os.path.exists(d):
                raise IOError('Directory does not exist: {}'.format(d))
            elif not os.access(d, os.W_OK):
                raise PermissionError('Cannot write to {}'.format(d))
    
    def _obs_config_asdict(self):
        d = {}
        for k, cfg in self.obs_config.items():
            as_dict = {}
            as_dict.update(**cfg)
            d[k] = as_dict
        return d
    
    def _model_config_asdict(self):
        d = {}
        for k, cfg in self.model_config.items():
            as_dict = {}
            as_dict.update(**cfg)
            d[k] = as_dict
        return d
    
    def _set_obsconfig(self, val):
        cfg = {}
        for k, v in val.items():
            cfg[k] = ObsConfigEval(**v)
        self.obs_config = cfg
    
    def _set_modelconfig(self, val):
        cfg = {}
        for k, v in val.items():
            if not 'mtype' in v:
                raise KeyError('Need key "mtype" in specfication of model {}'
                               .format(k))
            cfg[k] = ModelConfigEval(**v)
        self.model_config = cfg

    def _save_map_json(self, map_data, obs_name, obs_var, vert_which,
                       mod_name=None, mod_var=None):
        map_outname = ('OBS-{}:{}_{}_MOD-{}:{}.json'.format(obs_name, 
                                                            obs_var,
                                                            vert_which,
                                                            mod_name,
                                                            mod_var))
        
        outfile_map =  os.path.join(self.out_dirs['map'], map_outname)
        with open(outfile_map, 'w') as f:
            simplejson.dump(map_data, f, ignore_nan=True)
        return map_outname
    
    def _save_stat_json(self, trends_stat, obs_var, obs_name, vert_which,
                        mod_name=None, mod_var=None):
        """Save station json file
        
        Parameters
        ----------
        trends_stat 
            
        var_name
        
        network_id
        
        vert_which
            
        """
        station_name = trends_stat['station']
        filename = ('OBS-{}:{}_{}_MOD-{}:{}_{}.json'.format(obs_name, 
                                                            obs_var,
                                                            vert_which,
                                                            mod_name,
                                                            mod_var,
                                                            station_name))
        
        outfile = os.path.join(self.out_dirs['ts'], filename)
        try:
            df = pd.DataFrame(trends_stat)
            df.T.reset_index().to_json(outfile, double_precision=5) 
        except ValueError as e:        
            raise ValueError('FATAL: could not save station time-series trends '
                             'data to json. Reason: {}'.format(repr(e)))
        return filename
    
    def _remove_outliers(self, stat, var_name, min_max, logfile=None):
        """Remove outliers from StationData objects
        
        NOTE
        ----
        This method is deprecated since 26.6.2019, since outlier removal is now
        done already in UngriddedData object.
        """
        
        raw = stat[var_name]
        len0 =  len(raw)
        notnan = ~np.isnan(raw)
        
        vals = raw[notnan]
        
        notok = np.logical_or(vals<min_max[0], vals>min_max[1])
        #notok = ~((vals > min_max[0]) & (vals < min_max[1]))
        vals_invalid = vals[notok]
        
        if len(vals_invalid) == 0:
            return stat
        
        if logfile is not None:        
            for dtime, val in vals_invalid.iteritems():
                logfile.write('{},{},{},{},{:.3f}\n'.format(stat.dataset_name, 
                                                        stat.station_name, 
                                                        var_name,
                                                        dtime,
                                                        val))
                
        try:
            stat[var_name].ix[vals_invalid.index] = np.nan
        except:
            pass
        
        if not len(stat[var_name]) == len0:
            raise Exception('Length mismatch of input and output arrays, developers, please check')
        elif all(np.isnan(stat[var_name])):
            raise DataCoverageError('No valid data remains after removing outliers')
        return stat
    
    def _station_to_timeseries(self, station, var_name, **alt_range):
        # load additional information about data source (if applicable)
        station.load_dataset_info()
        keymap = self.KEYMAP
        vardata = {}
        try:
            vardata['data_revision'] = station.data_revision
        except:
            vardata['data_revision'] = None
        vardata['pyaerocom_version'] = __version__
        vardata['data_overlap'] = False
        vardata['var_name'] = var_name
        
        if var_name in station.overlap:
            vardata['data_overlap'] = True
        if var_name in station['var_info']:
            var_info =  station['var_info'][var_name]
        else:
            var_info = {}
        for k, v in keymap.items():
            val = None
            if v in ['var_name', 'data_overlap']:
                continue
            elif v in vardata:
                raise Exception('Key {} already exists... please debug...'
                                .format(k))
            try: # longitude, latitude and altitude are @property decorators in StationData
                val = station[k]
            except:   
                if k in var_info:
                    val = var_info[k]
            if isinstance(val, (list, tuple)):
                val = '; '.join([str(x) for x in val])
            vardata[v] = val
        
        # this is necessary, as e.g. altitude attribute of input station data 
        # is not necessarily the station coordinate!
        for cname, cval in station.get_station_coords().items():
            if not isnumeric(cval):
                raise ValueError
                
            vardata[keymap[cname]] = np.float64(cval)
        # the actual sampling frequency
        freq = vardata['freq']
        if vardata['wavelength'] is None:
            try:
                wvl = '{} nm'.format(const.VARS[var_name].wavelength_nm)
                vardata['wavelength'] = wvl
            except:
                pass
        ts_types = self.TS_TYPES
        if freq in ts_types:
            #raise AttributeError('Invalid temporal resolution code {}'.format(freq))
            if ts_types.index(freq) <= ts_types.index('daily'):
                to_freq = 'daily'
                freq_name = 'dobs'
            elif ts_types.index(freq) <= ts_types.index('monthly'):
                to_freq = 'monthly'
                freq_name = 'mobs'
            else:
                raise NotImplementedError
        elif freq == 'n/d':
            to_freq = 'daily'
            freq_name = 'dobs'
        else:
            raise AttributeError('Invalid ts_type {} in data...'.format())
        
        s = station.to_timeseries(var_name, freq=to_freq,
                                  resample_how=self.avg_how, 
                                  **alt_range)
        
        if len(s) == 0 or all(np.isnan(s)):
            raise DataCoverageError('Failed to retrieve timeseries')
        values = s.values
        index = s.index.values
        
        jsdate = self._to_jsdate(index)
        dates = pd.DatetimeIndex(index)
        
        
        d = od(year=dates.year,
               month=dates.month,
               day=dates.day,
               value=values,
               date=index,
               jsdate=jsdate)    
    
        vardata[freq_name] = pd.DataFrame(d)
        
        return vardata
    
    def _daily2monthly(self, daily, min_dim):
        """Helper to convert daily to monthly 
        """
        # Group data first by year, then by month
        g = daily.groupby(["year", "month"])
        # For each group, calculate the average of value
        _m = g.aggregate({"value":np.mean})
        numdays = g.size()
        
        # NOTE: BEFORE 16.5.19
        #invalid_mask = numdays <= min_dim
        # NOTE: AFTER 16.5.19
        invalid_mask = numdays < min_dim
        _m['value'].loc[invalid_mask] = np.nan
        
        
        #js date
        _m.reset_index(inplace=True)
        if not len(_m['value']) > 0:
            raise DataCoverageError('Derived monthly averages do not contain '
                                    'data')
        # Todo: check this, this seems cumbersome
        dates = _m.apply(lambda row: datetime.datetime(int(row['year']),
                                                       int(row['month']), 
                                                       15), axis=1)
     
        
        monthly = pd.Series(_m.value.values, dates.values)#.resample('MS').mean()
        monthly = monthly.resample('MS').mean()
        monthly.index = monthly.index.shift(14, 'D')
        
        dates = monthly.index
        jsdate = self._to_jsdate(monthly.index.values)
        
        return pd.DataFrame(od(year=dates.year,
                               month=dates.month,
                               day=dates.day,
                               value=monthly.values,
                               date=dates,
                               jsdate=jsdate))
    
    def _check_frequency_and_prep_mobs(self, data, min_dim):
        """Check frequency in data and if higher than monthly, create and append monthly
        
        Parameters
        ----------
        data : dict
            return value from :func:`_station_to_timeseries`
            
        Returns
        -------
        tuple
            2-element tuple, containing
            
            - str, granularity (daily, monthly)
            - DataFrame, containing monthly results
        """
        if 'dobs' in data:
            granulo = 'daily'
    
            #monthly avg
            #drop na values
            # FOLLOWING LINE COMMENTED OUT BY JGLISS (17.12.2018)
            daily = data['dobs'].dropna(subset=['value'])
            if len(daily) == 0:
                raise DataCoverageError('{}: Derived daily averages do not contain '
                                        'data'.format(data['station']))
            
            mobs = self._daily2monthly(daily, min_dim)
            data['mobs'] = mobs
            
        elif 'mobs' in data:
            granulo = 'monthly'
            mobs = data['mobs'].reset_index()
        else:
            raise Exception('Unexpected Error. Data should either contain daily '
                            'or monthly averages, please check function '
                            '_station_to_timeseries and debug')
        return (granulo, data)
    
    @staticmethod
    def _to_jsdate(dates):
        """Convert datetime vector to jsdate vector"""
        epoch = np.datetime64('1970-01-01')
        
        return (dates - epoch).astype('timedelta64[ms]').astype(int)
    
    @staticmethod
    def _init_trends_result_dict(start_yr):
        return _init_trends_result_dict(start_yr)
    
    @staticmethod
    def _compute_trend_error(m, m_err, v0, v0_err):
        """Computes error of trend estimate using gaussian error propagation
        
        The (normalised) trend is computed as T = m / v0
        
        where m denotes the slope of a regression line and v0 denotes the 
        normalistation value. This method computes the uncertainty of T (delta_T)
        using gaussian error propagation of uncertainties accompanying m and v0.
        
        Parameters
        ----------
        m : float
            slope in units of <U> yr-1 (where <U> denotes the unit of the data).
            (m -> "montant").
        m_err : float
            slope error (same unit as `m`)
        v0 : float
            normalisation value in units of <U>
        v0_err : float
            error of `v0` (same units as `v0`)
        
        Returns
        -------
        float
            error of T in computed using gaussian error propagation of trend 
            formula in units of %/yr
        """
        return _compute_trend_error(m, m_err, v0, v0_err)
    
    @staticmethod
    def _compute_trend_error_alt(reg, residual_magnitude, dt):
        """Alternative way to compute error of trend estimate 
        
        Parameters
        ----------
        m : float
            slope in units of <U> yr-1 (where <U> denotes the unit of the data).
            (m -> "montant").
        
        Returns
        -------
        float
            error of T computed using gaussian error propagation of trend formula
        """
        v0 = reg[0]
        vlast = reg[-1]
        v0_low = v0 - residual_magnitude
        v0_high = v0 + residual_magnitude
        
        vlast_low = vlast - residual_magnitude
        vlast_high = vlast + residual_magnitude
        
        T1 = (vlast_low - v0_high) / (dt * v0_high) *100
        T2 = (vlast_high - v0_low) / (dt * v0_low) * 100
        
        return (T1, T2)
    
    def _compute_trends_helper(self, data, periods, seasons):
        #trends with yearly and seasonal averages
         #add season to monthly data
        mobs = data['mobs']
        
        mobs['season'] = mobs.apply(lambda row: _get_season(row['month'],
                                                            row['year']), 
                                    axis=1)
        
        mobs = mobs.dropna(subset=['value'])
        # FOLLOWING LINE COMMENTED OUT BY JGLISS (17.12.2018)
        #data['dobs'].dropna(subset=['value'],inplace=True)
        # Group data first by year, then by month
        
        yrs = np.unique(mobs['year'])
        
        for i, seas in enumerate(seasons):
            dates = []
            
            #initialize seasonal object
            data[seas] = {'date'    : [], 
                          'jsdate'  : [], 
                          'val'     : [],
                          'trends'  : {}}
            #filter the months
            for yr in yrs:
                if seas == 'all': #yearly trends
                    catch = mobs[mobs['year'] == yr]
                else:
                    catch = mobs[mobs['season'].str.contains('{}-{}'.format(seas, yr))]
                
                dates.append(_mid_season(seas, yr))
    
                #needs 4 seasons to compute seasonal average to avoid biases
                if seas=='all' and len(np.unique(catch['season'].values)) < 4:
                    data[seas]['val'].append(np.nan)
                else:
                    data[seas]['val'].append(np.nanmean(catch['value']))
            # assign dates and jsdates vector to data dict
            data[seas]['date'] = np.asarray(dates)
            if len(dates) > 0:
                data[seas]['jsdate'] = self._to_jsdate(data[seas]['date'])
            else:
                data[seas]['jsdate'] = np.asarray([])
            
            #filter period
            for period in periods:
                # init dictionary that will contain trends results for this period
                data[seas]['trends'][period] = {}
                
                # desired start / stop year (note, that this may change if first 
                # or last value in tseries (or both) is NaN)
                start_yr, stop_yr = _years_from_periodstr(period)
                
                start_date = _mid_season(seas, start_yr)
                stop_date = _mid_season(seas, stop_yr)
                
                period_index = pd.date_range(start=start_date,
                                             end=stop_date, 
                                             freq=pd.DateOffset(years=1))
                
                num_dates_period = period_index.values.astype('datetime64[Y]').astype(np.float64)
                
                #filtering to the period limit
                jsp0 = self._to_jsdate(np.datetime64('{}-01-01'.format(start_yr)))
                jsp1 = self._to_jsdate(np.datetime64('{}-12-31'.format(stop_yr)))
                
                # vector containing numerical timestamps in javascript format
                jsdate = data[seas]['jsdate']
                dates = data[seas]['date']
                
                # get period filter mask
                tmask = np.logical_and(jsdate>=jsp0, jsdate<=jsp1) 
                
                # apply period mask to jsdate vector and value vector
                jsdate_data = jsdate[tmask]
                dates_data = dates[tmask]
                
                # vector containing data values
                vals = np.asarray(data[seas]['val'])[tmask]
                
                valid = ~np.isnan(vals)
                
                #works only on not nan values
                jsdate_data = jsdate_data[valid].astype(np.float64)
                dates_data = dates_data[valid]
                
                num_dates_data = dates_data.astype('datetime64[Y]').astype(np.float64)
                
                vals = vals[valid]           
                
                result = self._init_trends_result_dict(start_yr)
                
                #TODO: len(y) is number of years - 1 due to midseason averages
                result['n'] = len(vals)
                
                if len(vals) > 2:
                    result['y_mean'] = np.nanmean(vals)
                    result['y_min'] = np.nanmin(vals)
                    result['y_max'] = np.nanmax(vals)
                    
                    #Mann / Kendall test
                    [tau, pval] = kendalltau(x=num_dates_data, y=vals)
                    
                    
                    (slope, 
                     yoffs, 
                     slope_low, 
                     slope_up) = theilslopes(y=vals, x=num_dates_data, 
                                             alpha=self.slope_alpha)
                    
                    # estimate error of slope at input confidence level
                    slope_err = np.mean([abs(slope - slope_low), 
                                         abs(slope - slope_up)])
                    
                    reg_data = slope * num_dates_data + yoffs
                    reg_period = slope * num_dates_period  + yoffs
                    
                    
                    # value used for normalisation of slope to compute trend T
                    # T=m / v0
                    v0_data = reg_data[0]
                    v0_period = reg_period[0]
                    
                    # Compute the mean residual value, which is used to estimate
                    # the uncertainty in the normalisation value used to compute
                    # trend
                    mean_residual = np.mean(np.abs(vals - reg_data))
                    
                    # trend is slope normalised by first reference value. 
                    # 2 trends are computed, 1. the trend using the first value of
                    # the regression line at the first available data year, 2. the
                    # trend corresponding to the value corresponding to the first
                    # year of the considered period.
                    
                    trend_data = slope / v0_data * 100
                    trend_period =  slope / v0_period * 100
                    
                    # Compute errors of normalisation values
                    v0_err_data = mean_residual
                    t0_data, tN_data = num_dates_data[0], num_dates_data[-1]
                    t0_period = num_dates_period[0]
                    
                    # sanity check
                    assert t0_data < tN_data
                    assert t0_period <= t0_data
                    
                    dt_ratio = (t0_data - t0_period) / (tN_data - t0_data)
                    
                    v0_err_period = v0_err_data * (1 + dt_ratio)
                    
                    trend_data_err = self._compute_trend_error(m=slope,
                                                               m_err=slope_err,
                                                               v0=v0_data,
                                                               v0_err=v0_err_data)
                                                              
                    trend_period_err = self._compute_trend_error(m=slope,
                                                                 m_err=slope_err,
                                                                 v0=v0_period,
                                                                 v0_err=v0_err_period)
                                    
                    result['pval'] = pval
                    result['m'] = slope
                    result['m_err'] =slope_err
                    
                    result['slp'] = trend_data
                    result['slp_err'] = trend_data_err
                    result['reg0'] = v0_data
                    result['t0'] = jsdate_data[0]
                    if v0_period > 0:
                        result['slp_{}'.format(start_yr)] = trend_period
                        result['slp_{}_err'.format(start_yr)] = trend_period_err
                        result['reg0_{}'.format(start_yr)] = v0_period
                
                data[seas]['trends'][period] = result
        return data
    
    def _compute_trends_station(self, data, min_dim):
        """Compute trends for station
        
        Slightly modified code from original trends interface developed by 
        A. Mortier.
        
        Main changes applied:
            
            - Keep NaNs
        """
        granulo, data = self._check_frequency_and_prep_mobs(data, min_dim)
        
        seasons = self.SEASONS
        # get monthly values for display in interface ('mobs' is modified in 
        # _compute_trends_helper, e.g. NaNs are removed from it)
        jsdate = data['mobs'].jsdate.values
        monthly_vals = data['mobs'].value.values
        
        data = self._compute_trends_helper(data, self.periods, seasons)        
    
        #creates dictionary
        trends_stat = {'daily': {}, 'monthly': {}, 'season': {}}
        
        trends_stat['data_revision'] = data['data_revision']
        trends_stat['pyaerocom_version'] = data['pyaerocom_version']
        metakeys = self.KEYMAP.values()
        for metakey in metakeys:
            if metakey in data:
                trends_stat[metakey] = data[metakey]
        
        if granulo=='daily':
            trends_stat['daily']['date'] = data['dobs']['jsdate'].values
            trends_stat['daily']['val'] =  data['dobs']['value'].values
        elif granulo=='monthly':
            trends_stat['daily']['date'] = []
            trends_stat['daily']['val'] =  []
            
        trends_stat['monthly']['date'] = jsdate
        trends_stat['monthly']['val'] =  monthly_vals
        for seas in seasons:
            trends_stat['season'][seas] = {'date'  : data[seas]['jsdate'], 
                                           'val'   : data[seas]['val'], 
                                           'trends': data[seas]['trends']}
        
        reg = _find_area(data['lat'], data['lon'], 
                         regions_dict=self.DEFAULT_REGIONS)
        if reg == 'WORLD':
            reg = None
        # station information for map view
        map_stat = {'site'      : data['station'], 
                    'lat'       : data['lat'], 
                    'lon'       : data['lon'], 
                    'alt'       : data['alt'], 
                    'region'    : reg}
        
        
        for seas in seasons:
            map_stat[seas] = data[seas]['trends']
    
        return trends_stat, map_stat
    
    def _init_logfiles(self, obs_name):
        from datetime import datetime
        nowstr = datetime.strftime(datetime.now(), '%Y%m%d')
        
        outlier_log = open(os.path.join(self.logdir,
                                        'Run_{}_outliers_{}.log'
                                        .format(obs_name, nowstr), 'w+'))
        err_log = open(os.path.join(self.logdir,
                                        'Run_{}_errors_{}.log'
                                        .format(obs_name, nowstr), 'w+'))
        files_log = open(os.path.join(self.logdir,
                                        'Run_{}_outfiles_{}.log'
                                        .format(obs_name, nowstr), 'w+'))
        return (outlier_log, err_log, files_log)
    
    def _run_single_2d(self, ungridded_data, vars_to_retrieve,
                       min_dim, models, name, vert_which, files_created, 
                       err_log):
    
        files_created = self._run_single_helper(ungridded_data=ungridded_data, 
                                                vars_to_retrieve=vars_to_retrieve, 
                                                name=name, 
                                                vert_which=vert_which,
                                                files_created=files_created,
                                                min_dim=min_dim, 
                                                models=models,
                                                err_log=err_log)
        return files_created
    
    def _run_single_3d(self, ungridded_data, vars_to_retrieve,
                       min_dim, models, name, files_created, err_log):
        if models is not None:
            raise NotImplementedError('Cannot yet colocate 3D observations '
                                      'with model data ...')
        for add_name, alt_range in self.VERT_LAYERS.items():
            files_created = self._run_single_helper(ungridded_data=ungridded_data, 
                                                    vars_to_retrieve=vars_to_retrieve, 
                                                    min_dim=min_dim,  
                                                    models=models,
                                                    name=name, 
                                                    vert_which=add_name,
                                                    files_created=files_created,
                                                    err_log=err_log,
                                                    altitude=alt_range)
        return files_created
    
    def _run_single_helper(self, ungridded_data, vars_to_retrieve,
                           name, vert_which, files_created, min_dim, 
                           models, err_log=None, **alt_range):
        from pyaerocom.exceptions import DataCoverageError
        
        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
            
        model_access = {}
        if models is not None and len(models) > 0:
            model_access = self.check_model_access(models,
                                                   vars_to_retrieve)
              
        for var in vars_to_retrieve:
            
            all_stats = ungridded_data.to_station_data_all(vars_to_convert=var,
                                                           merge_if_multi=True,
                                                           by_station_name=True)
            stations = all_stats['stats']
            
            var_out = var
            stats_processed = []
            map_data = []
            
            stats_ok = []
            lats_ok = []
            lons_ok = []
            
            for stat in stations:
                stat_name = stat['station_name']
                print(var, stat['station_name'], name)
                if stat_name in stats_processed:
                    raise Exception('Unexpected Error, please check')
                
                try:
                    data_dict = self._station_to_timeseries(stat, 
                                                            var_name=var, 
                                                            **alt_range)
                    (trends_stat, 
                     map_stat) = self._compute_trends_station(data_dict,
                                                              min_dim=min_dim)
        
                    fname = self._save_stat_json(trends_stat, 
                                                 obs_var=var_out, 
                                                 obs_name=name,
                                                 vert_which=vert_which)
                    files_created['ts'].append(fname)
                           
                    stats_ok.append(stat_name)
                    lats_ok.append(stat['latitude'])
                    lons_ok.append(stat['longitude'])
                    if map_stat:
                        map_data.append(map_stat)
                        
                except DataCoverageError as e:
                    msg = 'Error: {} {}, {}'.format(var, stat.station_name,repr(e))
                    print(msg)
                    if err_log:
                        err_log.write(msg + '\n')
                
                
         
            # Add to map only if at least one station is available
            if len(stats_ok) > 0:
                map_outname = self._save_map_json(map_data, obs_name=name, 
                                                  obs_var=var_out,
                                                  vert_which=vert_which)
                files_created['map'].append(map_outname)
                
            for mod_name, mod_info in model_access.items():
                if not var in mod_info['vars_avail']:
                    continue
                mod_var=mod_info['vars_avail'][var]
                self._run_gridded(obs_name=name, obs_var=var, 
                                  mod_name=mod_name, mod_var=mod_var,
                                  stat_lats=lats_ok,
                                  stat_lons=lons_ok,
                                  stat_names=stats_ok, 
                                  min_dim=min_dim,
                                  vert_which=vert_which,
                                  files_created=files_created, 
                                  err_log=err_log,
                                  **alt_range)
        return files_created
    
    def freq_gridded(self):
        """Frequency used for any colocated gridded data"""
        return self.GRIDDED_TS_TYPE
    
    def _run_gridded(self, obs_name, obs_var, mod_name, mod_var, stat_lats, 
                     stat_lons, stat_names, min_dim, vert_which, files_created, 
                     err_log=None, **alt_range):
        mcfg = self.model_config[mod_name]
        data_id = mcfg['model_id']
        
        reader = ReadGridded(data_id=data_id)
        
        rf = mcfg['model_ts_type_read']
        
        read_freq = self.freq_gridded()
        if rf is not None:
            if isinstance(rf, str):
                read_freq = rf
            elif isinstance(rf, dict) and mod_var in rf:
                    read_freq = read_freq[mod_var]
        
        data = reader.read_var(mod_var, ts_type=read_freq,
                               ts_type_flex=True)
        
        data = data.resample_time(to_ts_type=self.freq_gridded())
        stations = data.to_time_series(longitude=stat_lons,
                                       latitude=stat_lats,
                                       add_meta=dict(station_name = stat_names))
        
        stats_processed = []
        map_data = []
        
        got_one = False
        for stat in stations:
            stat_name = stat['station_name']
            print(mod_var, stat['station_name'], obs_name, mod_name)
            if stat_name in stats_processed:
                raise Exception('Unexpected Error, please check')
            
            try:
    
                data_dict = self._station_to_timeseries(stat, 
                                                        var_name=mod_var, 
                                                        **alt_range)
                (trends_stat, 
                 map_stat) = self._compute_trends_station(data_dict,
                                                          min_dim=min_dim)
    
                fname = self._save_stat_json(trends_stat, 
                                             obs_var=obs_var, 
                                             obs_name=obs_name,
                                             vert_which=vert_which,
                                             mod_name=mod_name, 
                                             mod_var=mod_var)
                files_created['ts'].append(fname)
                got_one = True        
                    
                if map_stat:
                    map_data.append(map_stat)
                    
            except DataCoverageError as e:
                msg = ('Error: {} {}, {}'.format(mod_var, stat.station_name,
                                                 repr(e)))
                print(msg)
                if err_log:
                    err_log.write(msg + '\n')
                
                
         
        # Add to map only if at least one station is available
        if got_one:
            map_outname = self._save_map_json(map_data, 
                                              obs_name=obs_name, 
                                              obs_var=obs_var,
                                              vert_which=vert_which,
                                              mod_name=mod_name, 
                                              mod_var=mod_var)
            files_created['map'].append(map_outname)
        
        return files_created

def check_ebas_default():
    """Make sure the EBAS default configuration has not changed"""
    from pyaerocom.io import ReadEbas
    r = ReadEbas()
    assert r.opts.eval_flags
    assert r.opts.keep_aux_vars == False
            
if __name__ == '__main__':    
    t = TrendsEvaluation()
    
    